/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Annozilla.
 *
 * The Initial Developer of the Original Code is
 * Doug Daniels <rainking@alumni.rice.edu>
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Kathleen Brade <brade@comcast.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * Constants.
 */
const NS_ANNOTATION_CID = Components.ID("{750d793d-7958-4162-a70d-daff71f0bc88}");
const NS_ANNOTATION_PROG_ID = "@mozilla.org/annotation;1";

function Annotation() {
  this.mData = null;
  this.mServerName = null;
  this.mAnnotationURI = null;
  this.mBodyURI = null;
  this.mEncoding = null;
  this.bodyText = null;
  this.initialized = false;
}

Annotation.prototype = {

    get data() { return this.mData; },
    set data() { throw Components.Exception("You cannot set the metadata."); },
    
    get serverName() { return this.mServerName; },
    set serverName(name) { this.mServerName = name; },

    get annotationURI() { return this.mAnnotationURI; },
    set annotationURI(uri) { this.mAnnotationURI = uri; },

    get bodyURI() { return this.mBodyURI; },
    set bodyURI(uri) { this.mBodyURI = uri; },

    get encoding() { return this.mEncoding; },
    set encoding(ndo) { this.mEncoding = ndo; }
};

/**
 * Must be called after
 * createInstance to initialize 
 * this annotation.
 *
 * @param serverName
 * @param annotationURI
 * @param bodyURI
 */
Annotation.prototype.initialize = function(name, annotURI, bodyURI, encoding) {
    if (this.initialized) {
	throw Components.results.NS_ERROR_ALREADY_INITIALIZED;
    }

    this.mData = Components.classes["@mozilla.org/annotation-metadata;1"].createInstance(Components.interfaces.nsIAnnotationMetadata);
    
    this.mServerName = name;
    this.mAnnotationURI = annotURI;
    this.mBodyURI = bodyURI;
    this.mEncoding = encoding;

    this.initialized = true;
}

Annotation.prototype.QueryInterface = function(iid) {
  if (!iid.equals(Components.interfaces.nsIAnnotation) &&
      !iid.equals(Components.interfaces.nsISupports)) {
	throw Components.results.NS_ERROR_NO_INTERFACE;
      }
  return this;
}

/**
 * Get the text of the body of an annotation.
 * If the body text has not been set by setBodyText
 * or loaded using the AnnotationsApplication method
 * loadAnnotationBody, an exception will be thrown.
 *
 * @return string text of the annotation's body or null if bodyText not defined
 */
Annotation.prototype.getBodyText = function() {
  if (! this.initialized) {
    throw Components.results.NS_ERROR_NOT_INITIALIZED;
  }
  
  if (this.bodyText) {
    return this.bodyText;
  }
  else {
    return null;
  }
}

/**
 * Set the text of the body of the annotation.
 *
 * @param text string 
 */
Annotation.prototype.setBodyText = function(text) {
    if (! this.initialized) {
	throw Components.results.NS_ERROR_NOT_INITIALIZED;
    }
    
    this.bodyText = text;
}

Annotation.prototype.toString = function() {
  return "nsIAnnotation (" + this.annotationURI + ")";
}

/**
 * Build a deep clone of the annotation, including
 * all Metadata
 * 
 * @return deep clone of Annotation
 */
Annotation.prototype.clone = function() {

  var clonedAnnot = new Annotation();

  clonedAnnot.mData = this.mData.clone();
  clonedAnnot.mServerName = this.mServerName;
  clonedAnnot.mAnnotationURI = this.mAnnotationURI;
  clonedAnnot.mBodyURI = this.mBodyURI;
  clonedAnnot.initialized = this.initialized;

  if (this.bodyText) {
    clonedAnnot.bodyText = this.bodyText;
  }

  return clonedAnnot.QueryInterface(Components.interfaces.nsIAnnotation);
}
  
  

/************
 * XPCOM-ness
 ************/


var annotationModule = {
    firstTime: true,
    
    /*
     * RegisterSelf is called at registration time (component installation
     * or the only-until-release startup autoregistration) and is responsible
     * for notifying the component manager of all components implemented in
     * this module.  The fileSpec, location and type parameters are mostly
     * opaque, and should be passed on to the registerComponent call
     * unmolested.
     */
    registerSelf: function (compMgr, fileSpec, location, type) {
        if (this.firstTime) {
            dump("*** Deferring registration of nsAnnotation\n");
            this.firstTime = false;
            throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
        }
        debug("*** Registering nsAnnotation\n");
        dump("*** Registering nsAnnotation\n");
        compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
        compMgr.registerFactoryLocation(NS_ANNOTATION_CID,
                                        "Annotation",
                                        NS_ANNOTATION_PROG_ID, 
                                        fileSpec,
                                        location,
                                        type);
    },

    /*
     * The GetClassObject method is responsible for producing Factory and
     * SingletonFactory objects (the latter are specialized for services).
     */
    getClassObject: function (compMgr, cid, iid) {
        if (!cid.equals(NS_ANNOTATION_CID)) {
	    throw Components.results.NS_ERROR_NO_INTERFACE;
	}

        if (!iid.equals(Components.interfaces.nsIFactory)) {
	    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
	}

        return this.annotationFactory;
    },

    /* factory object */
    annotationFactory: {
        /*
         * Construct an instance of the interface specified by iid, possibly
         * aggregating it with the provided outer.  (If you don't know what
         * aggregation is all about, you don't need to.  It reduces even the
         * mightiest of XPCOM warriors to snivelling cowards.)
         */
        createInstance: function (outer, iid) {
            if (outer != null) {
		throw Components.results.NS_ERROR_NO_AGGREGATION;
	    }

            return (new Annotation()).QueryInterface(iid);
        }
    },

    /*
     * The canUnload method signals that the component is about to be unloaded.
     * C++ components can return false to indicate that they don't wish to be
     * unloaded, but the return value from JS components' canUnload is ignored:
     * mark-and-sweep will keep everything around until it's no longer in use,
     * making unconditional ``unload'' safe.
     *
     * You still need to provide a (likely useless) canUnload method, though:
     * it's part of the nsIModule interface contract, and the JS loader _will_
     * call it.
     */
    canUnload: function(compMgr) {
        return true;
    }
};

function NSGetModule(compMgr, fileSpec) {
    return annotationModule;
}
